import io
import os
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import httpx
from loguru import logger

CURRENT_FILE_PATH = Path(__file__).resolve()
DB_GENERIC_PATH = CURRENT_FILE_PATH.parent.parent.parent / "db"
DB_FOLDER_NAME = "{time}"
CARD_DB_DOWNLOAD_URL = "https://mtgjson.com/api/v5/AllPrintings.sqlite.zip"
AUTO_UPDATE_DB = True
AUTO_UPDATE_DB_INTERVAL = 24 * 7  # in hours

logger.disable("mtgcdb")


def download_card_definitions_db() -> Path:
    """Download card definitions DB"""
    logger.info(f"Downloading card definitions DB from {CARD_DB_DOWNLOAD_URL}")
    response = httpx.get(CARD_DB_DOWNLOAD_URL)
    if response.status_code != 200:
        logger.error(
            f"Failed to download card definitions DB. Status code: {response.status_code}"
        )
        raise Exception("Failed to download card definitions DB")

    zip_db = ZipFile(io.BytesIO(response.content))
    current_time = datetime.now().strftime("%Y%m%dT%H%M%S")
    db_folder_path = DB_GENERIC_PATH / DB_FOLDER_NAME.format(time=current_time)
    db_folder_path.mkdir(parents=True, exist_ok=True)
    zip_db.extractall(path=db_folder_path)
    db_path = db_folder_path / "AllPrintings.sqlite"

    if os.path.exists(db_path):
        logger.success(f"Card definitions saved to {db_path}")
    else:
        raise Exception("Failed to extract card definitions DB")

    return db_path


def get_most_recent_db() -> Path | None:
    """Get the most recent card definitions DB"""
    DB_GENERIC_PATH.mkdir(parents=True, exist_ok=True)
    db_folders = os.listdir(DB_GENERIC_PATH)
    if not db_folders:
        return None

    db_folders.sort()
    latest_db_folder = db_folders[-1]
    latest_db_folder_path = DB_GENERIC_PATH / latest_db_folder
    db_files = os.listdir(latest_db_folder_path)
    if not db_files:
        return None

    db_files.sort()
    latest_db_file = db_files[0]
    latest_db_file_path = latest_db_folder_path / latest_db_file
    return latest_db_file_path


def clean_old_dbs() -> None:
    """Clean old card definitions DBs"""
    db_folders = os.listdir(DB_GENERIC_PATH)
    if not db_folders:
        return

    db_folders.sort()
    for db_folder in db_folders[:-1]:
        db_folder_path = DB_GENERIC_PATH / db_folder
        logger.info(f"Removing old card definitions DB: {db_folder_path}")
        # Remove the folder and its contents
        for file in db_folder_path.iterdir():
            file.unlink()
        db_folder_path.rmdir()


def update_or_pass() -> Path:
    """Check if the card definitions DB should be updated.
    Returns the path to the most recent DB. If no DB is found, download a new one."""

    db_path = get_most_recent_db()
    if not db_path:
        logger.info("No card definitions DB found")
        return download_card_definitions_db()

    latest_db_folder = db_path.parent.name
    db_file_time = datetime.strptime(latest_db_folder, "%Y%m%dT%H%M%S")
    time_diff = datetime.now() - db_file_time
    time_diff_hours = time_diff.total_seconds() / 3600

    if time_diff_hours >= AUTO_UPDATE_DB_INTERVAL and AUTO_UPDATE_DB:
        logger.info(
            f"Updating card definitions DB. Last updated {time_diff_hours/7:.2f} days ago"
        )
        clean_old_dbs()
        return download_card_definitions_db()
    else:
        logger.info(
            f"Skipping card definitions DB update. Last updated {time_diff_hours/7:.2f} days ago"
        )
        return db_path



from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path

from sqlalchemy import create_engine, or_, select
from sqlalchemy.orm import sessionmaker

from mtgcdb.card import Base, Card
from mtgcdb.db import clean_old_dbs, download_card_definitions_db, update_or_pass


class MTGCDB:
    """Magic: The Gathering card database singleton"""

    __instance: "MTGCDB | None" = None

    def __init__(self, force_update: bool = False):
        if (
            MTGCDB.__instance is not None
            and hasattr(self, "engine")
            and hasattr(self, "Session")
        ):
            return MTGCDB.__instance  # type: ignore

        self.db_path = update_or_pass() if not force_update else self.update_db()
        self.engine = create_engine(f"sqlite:///{self.db_path}", echo=False)
        self.Session = sessionmaker(bind=self.engine)
        Base.prepare(self.engine)

        MTGCDB.__instance = self

    def update_db(self) -> Path:
        clean_old_dbs()
        return download_card_definitions_db()

    @lru_cache(maxsize=1024)
    def get_card_by_name(self, name: str, set_code: str | None = None) -> Card:
        """Get a card by its name

        Args:
            name (str): exact card name
            set_code (str | None, optional): 3 letters of the set code. Defaults to None.

        Returns:
            Card:
        """
        with self.Session() as session:
            query = select(Card).where(Card.name == name)
            if set_code:
                query = query.where(Card.setCode == set_code)
            card = session.scalars(query).first()
            if not card:
                card = self.get_card_by_face_name(name, set_code=set_code)
            return card

    @lru_cache(maxsize=1024)
    def get_card_by_face_name(
        self, face_name: str, set_code: str | None = None
    ) -> Card:
        """Get a card by its front face name

        Args:
            face_name (str): exact front face name
            set_code (str | None, optional): 3 letters of the set code. Defaults to None.

        Returns:
            Card:
        """
        with self.Session() as session:
            query = select(Card).where(Card.faceName == face_name)
            if set_code:
                query = query.where(Card.setCode == set_code)
            card = session.scalars(query).first()
            if not card:
                raise ValueError(f"Card faceName={face_name} not found in database")
            return card

    def get_card_image_url_from_name(
        self, name: str, set_code: str | None = None, img_type: str = "normal"
    ) -> str:
        """Get the image URL of a card by its name

        Args:
            name (str): exact card name
            set_code (str | None, optional):  3 letters of the set code. Defaults to None.
            img_type (str, optional): One of "png", "normal", "large" or "small". Defaults to "normal".
              See https://scryfall.com/docs/api/images for more info.

        Returns:
            str: URL of the card image
        """
        c = self.get_card_by_name(name, set_code=set_code)
        img_format = "png" if img_type == "png" else "jpg"
        scryfall_id = c.identifier.scryfallId
        url = f"https://cards.scryfall.io/{img_type}/front/{scryfall_id[0]}/{scryfall_id[1]}/{scryfall_id}.{img_format}"
        return url

    def get_cards_by_names(self, names: list[str]) -> list[Card]:
        """Get a list of cards by their names

        Args:
            names (list[str]): List of exact card names

        Returns:
            list[Card]: List of cards
        """
        with self.Session() as session:
            query = select(Card).where(
                or_(Card.name.in_(names), Card.faceName.in_(names))
            )
            results = session.scalars(query).all()
        return _deduplicate_cards_by_name(results)

    @lru_cache(maxsize=8)
    def get_cards_by_set_code(self, setCode: str) -> list[Card]:
        """Get all cards from a set by its set code

        Args:
            setCode (str): Set code (3 letters)

        Returns:
            list[Card]: List of cards from the set
        """
        with self.Session() as session:
            query = select(Card).where(Card.setCode == setCode)
            results = session.scalars(query).all()
        return _deduplicate_cards_by_name(results)


def _deduplicate_cards_by_name(cards: Sequence[Card]) -> list[Card]:
    """Deduplicate a list of cards by their names

    Args:
        cards (Sequence[Card]): List of cards

    Returns:
        list[Card]: Deduplicated list of cards
    """
    card_names = set()
    dedup_cards = []
    for card in cards:
        if card.name not in card_names:
            dedup_cards.append(card)
            card_names.add(card.name)
    return dedup_cards



from typing import Optional

from sqlalchemy import ForeignKey
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Mapped, mapped_column, relationship

Base = automap_base()


class Card(Base):  # type: ignore
    __tablename__ = "cards"
    extend_existing = True

    uuid: Mapped[str] = mapped_column(primary_key=True)
    name: Mapped[str]
    faceName: Mapped[str]
    types: Mapped[str]
    subtypes: Mapped[str]
    setCode: Mapped[str]
    power: Mapped[Optional[str]]
    toughness: Mapped[Optional[str]]
    text: Mapped[str]
    manaValue: Mapped[float]
    manaCost: Mapped[str]
    colors: Mapped[str]
    colorIdentity: Mapped[str]
    rarity: Mapped[str]

    def __repr__(self) -> str:
        return f"Card(name={self.name}, types={self.types}, manaCost={self.manaCost})"

    identifier: Mapped["CardIdentifier"] = relationship(
        back_populates="card", viewonly=True, lazy="joined"
    )


class CardIdentifier(Base):  # type: ignore
    __tablename__ = "cardIdentifiers"

    uuid = mapped_column(ForeignKey("cards.uuid"), primary_key=True)
    scryfallId: Mapped[str]

    card: Mapped[Card] = relationship(back_populates="identifier", viewonly=True)
